import matplotlib.pyplot as plt
import numpy as np

class History:
    def __init__(self):
        self.ep_rewards = []
        self.aggr_ep_rewards = {'ep':[], 'avg':[], 'min':[], 'max':[], 'acc_landings':[]}

    def update_reward(self, reward):
        self.ep_rewards.append(reward)

    def update_correct_landing(self, land_acc):
        self.aggr_ep_rewards["acc_landings"].append(land_acc)

    def print_last(self, episode):
        print(f"Episode {episode} reward: {self.ep_rewards[-1]}")


    def print_average_last(self, episode, tail):
        average_reward = np.round(sum(self.ep_rewards[-tail:]) / tail, 4)
        self.aggr_ep_rewards['ep'].append(episode)
        self.aggr_ep_rewards['avg'].append(average_reward)
        self.aggr_ep_rewards['min'].append(np.round(min(self.ep_rewards[-tail:]), 4))
        self.aggr_ep_rewards['max'].append(np.round(max(self.ep_rewards[-tail:]), 4))
        print(f"Episode: {episode} | Reward avg: {average_reward} min: "
              f"{self.aggr_ep_rewards['min'][-1]} max: {self.aggr_ep_rewards['max'][-1]}"
              f" landing_accuracy: {np.round(self.aggr_ep_rewards['acc_landings'][-1]*100,2)}% (over last {tail} episodes)")



def plot_hist(rad, title):
    plt.hist(rad, 75, density=True, facecolor='g', alpha=0.75)
    plt.xlabel(title)
    plt.grid(True)
    plt.tight_layout()
    plt.savefig(f"hist_{title}.png", dpi=300)


def heatmap2d(agent, episode):
    arr = agent.Q_table.copy()
    arr.resize((6*7*2*2*4, 7*5*5*5))
    plt.imshow(arr, cmap='viridis')
    plt.colorbar()
    plt.axis('off')
    plt.title(f"Episode {episode} (LunarLander-v2 Q-table)")
    plt.savefig(f"animation/{episode}.png", dpi=145)
    plt.clf()

def MA(rad, window):
    #newr = np.convolve(rad, [1]*window, 'valid')/window
    newr = []
    for i in range(len(rad)+1 - window):
        k = 0
        for j in range(window):
            k += rad[i+j]
        newr.append(k/window)
    return np.array(newr)

def plot_experiments(exp):
    alphas = [0.05, 0.2, 0.4]
    gammas = [0.825, 0.9, 0.975]
    epsilons = [0.1, 0.5, 0.9]
    fig, axes = plt.subplots(3, 1, figsize=(15, 8), sharex=True)
    for i, a1 in enumerate(alphas):
        ax = axes[i]
        if i == 0:
            ax.set_title(r'Landing accuracy of last 100 episodes')
        ax.set_xlabel('episodes')
        for j, g1 in enumerate(gammas):
            for k, e1 in enumerate(epsilons):
                ax.plot(MA(exp[(a1, g1, e1)].aggr_ep_rewards['acc_landings'],1),linewidth=0.95,label=r'$\alpha:$'f'{a1} ' + r'$\gamma:$'f'{g1} ' + r'$\epsilon:$'f'{e1}')
        if True:
            ax.legend(bbox_to_anchor=(1.05, 1), loc='upper left', ncol=1)
    fig.subplots_adjust(hspace=0)
    fig.tight_layout()
    fig.subplots_adjust(hspace=0)
    plt.savefig('experiments_acc_landings.png', dpi = 400)


def plot_avg_reward():
    import matplotlib.pyplot as plt
    part1 = np.load('experiments/first/qlearning_part1/history_rewards_avg_part1.npy')
    part2 = np.load('experiments/first/qlearning_part2/history_rewards_avg_part2_episode_450000.npy')
    x = np.array(list(range(1, len(part1)+len(part2)+1)))*100000
    plt.plot(x, np.hstack((part1, part2)))
    plt.xlabel("episodes", fontsize=13)
    plt.ylabel("avg. Reward", fontsize=13)
    plt.title('Q-learning')
    plt.tight_layout()
    plt.savefig('avgrew.png', dpi=200)

def plot_avg_reward_hists():
    import matplotlib.pyplot as plt
    part1 = np.load('experiments/first_ql/qlearning_part1/history_ep_rewards_part1.npy')
    part2 = np.load('experiments/first_ql/qlearning_part2/history_ep_rewards_part2_episode_140000.npy')
    all = np.hstack((part1, part2))

    for i, episode in enumerate(range(5000, len(all), 5000)):
        plt.hist(all[episode-5000:episode], 75, density=True, facecolor='g', alpha=0.75)
        ax = plt.gca()
        ax.axes.yaxis.set_ticks([])
        plt.title(f'Distribution of rewars (last 5000) Episode: {episode}')
        plt.xlabel("Rewards", fontsize=13)
        plt.tight_layout()
        plt.savefig(f"gif/{i}.png", dpi=200)
        plt.clf()